/* Copyright (C) |Meso|Star> 2015-2017 (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s3d_c.h"
#include "s3d_instance.h"
#include "s3d_mesh.h"
#include "s3d_scene_c.h"

#include <rsys/float33.h>

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static int
check_primitive(const struct s3d_primitive* prim)
{
  return prim
      && prim->geom_id != S3D_INVALID_ID
      && prim->prim_id != S3D_INVALID_ID
      && prim->mesh__ != NULL
      && (prim->inst_id != S3D_INVALID_ID || prim->inst__ == NULL);
}

/*******************************************************************************
 * Exported functions
 ******************************************************************************/
res_T
s3d_primitive_get_attrib
  (const struct s3d_primitive* prim,
   const enum s3d_attrib_usage usage,
   const float uv[2],
   struct s3d_attrib* attrib)
{
  const uint32_t* ids;
  struct geometry* geom_mesh = NULL;
  const float* transform = NULL;
  char flip_surface = 0;
  float w;
  res_T res = RES_OK;

  if(!check_primitive(prim) || usage == S3D_ATTRIBS_COUNT__ || !uv || !attrib)
    return RES_BAD_ARG;

  /* Unormalized barycentric coordinates */
  w = CLAMP(1.f - uv[0] - uv[1], 0.f, 1.f);
  if(uv[0] < 0.f || uv[1] < 0.f || uv[0] > 1.f || uv[1] > 1.f
  || !eq_epsf(w + uv[0] + uv[1], 1.f, 1.e-3f))
    return RES_BAD_ARG;

  if(prim->inst__ == NULL) {
    geom_mesh = (struct geometry*)prim->mesh__;
    flip_surface = geom_mesh->flip_surface;
  } else {
    const struct geometry* geom_inst = (const struct geometry*)prim->inst__;
    ASSERT(geom_inst->type == GEOM_INSTANCE);
    ASSERT(prim->inst_id == geom_inst->name);
    geom_mesh = (struct geometry*)prim->mesh__;
    transform = geom_inst->data.instance->transform;
    ASSERT(geom_mesh);
    flip_surface = geom_inst->flip_surface ^ geom_mesh->flip_surface;
  }
  ASSERT(prim->geom_id == geom_mesh->name);
  ASSERT(geom_mesh->type == GEOM_MESH);

  /* The mesh haven't the required mesh attrib */
  if(usage != S3D_GEOMETRY_NORMAL && !geom_mesh->data.mesh->attribs[usage]) {
    res = RES_BAD_ARG;
    goto error;
  }

  /* Out of bound primitive index */
  if(prim->prim_id >= mesh_get_ntris(geom_mesh->data.mesh)) {
    res = RES_BAD_ARG;
    goto error;
  }
  ids = mesh_get_ids(geom_mesh->data.mesh) + prim->prim_id * 3/*#triangle ids*/;
  attrib->usage = usage;

  if(usage == S3D_POSITION || usage == S3D_GEOMETRY_NORMAL) {
    const float* v0, *v1, *v2;
    const float* pos;
    attrib->type = S3D_FLOAT3;
    /* Fetch data */
    pos = mesh_get_pos(geom_mesh->data.mesh);
    v0 = pos + ids[0] * 3;
    v1 = pos + ids[1] * 3;
    v2 = pos + ids[2] * 3;
    if(usage == S3D_GEOMETRY_NORMAL) { /* Compute the geometry normal */
      float e0[3], e1[3];
      /* Build the geometric normal with respect to surface orientation.
       * Default is Clock Wise */
      f3_sub(e0, v2, v0);
      f3_sub(e1, v1, v0);
      if(flip_surface) {
        f3_cross(attrib->value, e1, e0);
      } else {
        f3_cross(attrib->value, e0, e1);
      }
      if(transform) { /* Transform the normal from local to world space */
        float transform_invtrans[9];
        f33_invtrans(transform_invtrans, transform);
        f33_mulf3(attrib->value, transform_invtrans, attrib->value);
      }
    } else { /* Interpolate the vertex position */
      float tmp[3];
      f3_mulf(attrib->value, v0, uv[0]);
      f3_add(attrib->value, attrib->value, f3_mulf(tmp, v1, uv[1]));
      f3_add(attrib->value, attrib->value, f3_mulf(tmp, v2, w));
      if(transform) { /* Transform the position from local to world space */
        f33_mulf3(attrib->value, transform, attrib->value); /* Rotation */
        f3_add(attrib->value, attrib->value, transform + 9); /* Translation */
      }
    }
  } else {
    const float* attr;
    const float* v0, *v1, *v2;
    unsigned i, dim;
    attrib->type = geom_mesh->data.mesh->attribs_type[usage];
    /* Fetch attrib data */
    dim = s3d_type_get_dimension(attrib->type);
    attr = mesh_get_attr(geom_mesh->data.mesh, usage);
    v0 = attr + ids[0] * dim;
    v1 = attr + ids[1] * dim;
    v2 = attr + ids[2] * dim;
    /* Interpolate the vertex attribs */
    ASSERT(dim <= 4);
    FOR_EACH(i, 0, dim) {
      attrib->value[i] = v0[i]*uv[0] + v1[i]*uv[1] + v2[i]*w;
    }
  }
exit:
  return res;
error:
  goto exit;
}

res_T
s3d_primitive_has_attrib
  (const struct s3d_primitive* prim,
   const enum s3d_attrib_usage attr,
   char* has_attrib)
{
  if(!check_primitive(prim) || !has_attrib
  || (attr != S3D_GEOMETRY_NORMAL && (unsigned)attr >= S3D_ATTRIBS_COUNT__))
    return RES_BAD_ARG;

  if(attr == S3D_GEOMETRY_NORMAL) {
    *has_attrib = 1;
  } else {
    struct geometry* geom_mesh = (struct geometry*)prim->mesh__;
    ASSERT(geom_mesh->type == GEOM_MESH);
    *has_attrib = geom_mesh->data.mesh->attribs[attr] != NULL;
  }
  return RES_OK;
}

res_T
s3d_primitive_sample
  (const struct s3d_primitive *prim,
   const float u,
   const float v,
   float st[2])
{
  double sqrt_u;

  if(!check_primitive(prim) || !st)
    return RES_BAD_ARG;

  /* Expecting canonic numbers */
  if(u < 0.f || u >= 1.f || v < 0.f || v >= 1.f)
    return RES_BAD_ARG;

  /* Only triangular primitives are currently supported. So simply compute the
   * barycentric coordinates of an uniform triangle sampling. */
  sqrt_u = sqrt(u);
  st[0] = (float)(1.0 - sqrt_u);
  st[1] = (float)(v * sqrt_u);
  return RES_OK;
}

res_T
s3d_primitive_compute_area(const struct s3d_primitive* prim, float* area)
{
  const uint32_t* ids;
  const float* pos;
  const float* v0, *v1, *v2;
  float E0[3], E1[3], N[3];
  struct geometry* geom;

  if(!check_primitive(prim) || !area)
    return RES_BAD_ARG;

  geom = (struct geometry*)prim->mesh__;
  pos = mesh_get_pos(geom->data.mesh);
  ids = mesh_get_ids(geom->data.mesh) + prim->prim_id * 3/* #triangle ids */;
  v0 = pos + ids[0] * 3/* #coords */;
  v1 = pos + ids[1] * 3/* #coords */;
  v2 = pos + ids[2] * 3/* #coords */;
  f3_sub(E0, v1, v0);
  f3_sub(E1, v2, v0);
  *area = f3_len(f3_cross(N, E0, E1)) * 0.5f;
  return RES_OK;
}

res_T
s3d_primitive_get_transform
  (const struct s3d_primitive* prim, float transform[12])
{
  if(!check_primitive(prim) || !transform)
    return RES_BAD_ARG;

  if(!prim->inst__) {
    f3(transform + 0, 1.f, 0.f, 0.f);
    f3(transform + 3, 0.f, 1.f, 0.f);
    f3(transform + 6, 0.f, 0.f, 1.f);
    f3(transform + 9, 0.f, 0.f, 0.f);
  } else {
    struct geometry* geom_inst = (struct geometry*)prim->inst__;
    ASSERT(geom_inst->type == GEOM_INSTANCE);
    f3_set(transform + 0, geom_inst->data.instance->transform + 0);
    f3_set(transform + 3, geom_inst->data.instance->transform + 3);
    f3_set(transform + 6, geom_inst->data.instance->transform + 6);
    f3_set(transform + 9, geom_inst->data.instance->transform + 9);
  }
  return RES_OK;
}

res_T
s3d_triangle_get_vertex_attrib
  (const struct s3d_primitive* prim,
   const size_t ivertex,
   const enum s3d_attrib_usage usage,
   struct s3d_attrib* attrib)
{
  struct geometry* geom_mesh = NULL;
  const float* transform = NULL;
  const uint32_t* ids;

  if(!check_primitive(prim) || ivertex > 2
  || (unsigned)usage >=  S3D_ATTRIBS_COUNT__
  || !attrib) {
    return RES_BAD_ARG;
  }

  geom_mesh = (struct geometry*)prim->mesh__;
  ASSERT(prim->geom_id == geom_mesh->name);
  ASSERT(geom_mesh->type == GEOM_MESH);

  if(prim->inst__ != NULL) {
    const struct geometry* geom_inst = (const struct geometry*)prim->inst__;
    ASSERT(geom_inst->type == GEOM_INSTANCE);
    ASSERT(prim->inst_id == geom_inst->name);
    transform = geom_inst->data.instance->transform;
  }

  /* The mesh haven't the required mesh attrib */
  if(!geom_mesh->data.mesh->attribs[usage]) {
    return RES_BAD_ARG;
  }

  /* Out of bound primitive index */
  if(prim->prim_id >= mesh_get_ntris(geom_mesh->data.mesh)) {
    return RES_BAD_ARG;
  }
  ids = mesh_get_ids(geom_mesh->data.mesh) + prim->prim_id * 3/*#triangle ids*/;
  attrib->usage = usage;

  if(usage != S3D_POSITION) {
    const float* attr;
    unsigned i, dim;
    attrib->type = geom_mesh->data.mesh->attribs_type[usage];
    /* Fetch attrib data */
    dim = s3d_type_get_dimension(attrib->type);
    attr = mesh_get_attr(geom_mesh->data.mesh, usage) + ids[ivertex] * dim;
    FOR_EACH(i, 0, dim) attrib->value[i] = attr[i];
  } else {
    const float* pos;
    attrib->type = S3D_FLOAT3;
    /* Fetch data */
    pos = mesh_get_pos(geom_mesh->data.mesh) + ids[ivertex] * 3;
    f3_set(attrib->value, pos);
    if(transform) { /* Transform the position from local to world space */
      f33_mulf3(attrib->value, transform, attrib->value); /* Rotation */
      f3_add(attrib->value, attrib->value, transform + 9); /* Translation */
    }
  }
  return RES_OK;
}

