/* Copyright (C) |Meso|Star> 2016-2017 (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#include "s3dut.h"
#include "s3dut_mesh.h"

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static void
setup_cylinder_coords
  (double* coords,
   const double radius,
   const double height,
   const unsigned nslices,
   const unsigned nstacks)
{
  double step_theta;
  double step_z;
  size_t itheta, istack;
  size_t i = 0;
  ASSERT(coords && radius > 0 && height > 0 && nslices >= 3 && nstacks >= 1);

  /* Contour vertices */
  step_theta = 2*PI / (double)nslices;
  step_z = height / (double)nstacks;
  FOR_EACH(itheta, 0, nslices) {
    const double theta = (double)itheta * step_theta;
    const double x = cos(theta);
    const double y = sin(theta);
    double z = -height * 0.5;

    FOR_EACH(istack, 0, nstacks+1) {
      coords[i++] = x*radius;
      coords[i++] = y*radius;
      coords[i++] = z;
      z += step_z;
    }
  }

  /* Bottom polar vertex */
  coords[i++] = 0;
  coords[i++] = 0;
  coords[i++] = -height * 0.5;

  /* Top polar vertex */
  coords[i++] = 0;
  coords[i++] = 0;
  coords[i++] = height * 0.5;
}

static void
setup_cylinder_indices
  (size_t* ids,
   const unsigned nslices,
   const unsigned nstacks)
{
  size_t islice;
  size_t istack;
  size_t ibottom;
  size_t itop;
  size_t i = 0;
  ASSERT(ids && nslices && nstacks);

  FOR_EACH(islice, 0, nslices) {
    const size_t islice0 = islice * (nstacks + 1);
    const size_t islice1 = ((islice + 1) % nslices) * (nstacks + 1);
    FOR_EACH(istack, 0, nstacks) {
      const size_t istack0 = istack + 0;
      const size_t istack1 = istack + 1;

      ids[i++] = islice0 + istack0;
      ids[i++] = islice0 + istack1;
      ids[i++] = islice1 + istack0;

      ids[i++] = islice1 + istack0;
      ids[i++] = islice0 + istack1;
      ids[i++] = islice1 + istack1;
    }
  }

  ibottom = nslices * (nstacks+1);
  FOR_EACH(islice, 0, nslices) {
    ids[i++] = ibottom;
    ids[i++] = islice * (nstacks+1);
    ids[i++] = ((islice+1)%nslices) * (nstacks+1);
  }

  itop = ibottom + 1;
  FOR_EACH(islice, 0, nslices) {
    ids[i++] = itop;
    ids[i++] = ((islice+1)%nslices) * (nstacks+1) + nstacks;
    ids[i++] = islice * (nstacks+1) + nstacks;
  }
}

/*******************************************************************************
 * Exported functions
 ******************************************************************************/
res_T
s3dut_create_cylinder
  (struct mem_allocator* allocator,
   const double radius,
   const double height,
   const unsigned nslices,
   const unsigned nstacks,
   struct s3dut_mesh** mesh)
{
  struct s3dut_mesh* cylinder = NULL;
  double* coords = NULL;
  size_t* ids = NULL;
  size_t nverts;
  size_t ntris;
  res_T res = RES_OK;

  if(radius <= 0 || height <= 0 || nslices < 3 || nstacks < 1 || !mesh) {
    res = RES_BAD_ARG;
    goto error;
  }

  nverts = nslices * (nstacks+1) /*#contour vers*/ + 2/*#polar verts*/;
  ntris = 2*nslices*nstacks /*#contour tris*/ + 2*nslices/* #cao tris*/;

  res = mesh_create(allocator, S3DUT_MESH_CYLINDER, nverts, ntris, &cylinder);
  if(res != RES_OK) goto error;

  coords = darray_double_data_get(&cylinder->coords);
  ids = darray_size_t_data_get(&cylinder->ids);
  setup_cylinder_coords(coords, radius, height, nslices, nstacks);
  setup_cylinder_indices(ids, nslices, nstacks);

exit:
  if(mesh) *mesh = cylinder;
  return res;
error:
  if(cylinder) {
    S3DUT(mesh_ref_put(cylinder));
    cylinder = NULL;
  }
  goto exit;
}

